#!/usr/bin/env python3

from abc import ABC, abstractmethod
from scipy.spatial import distance_matrix
import numpy as np


# Abstract kernel
class Kernel(ABC):
    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def eval(self, x, y):
        pass

    def eval_prod(self, x, y, v, batch_size=100):
        N = x.shape[0]
        num_batches = int(np.ceil(N / batch_size))
        mat_vec_prod = np.zeros((N, 1))
        for idx in range(num_batches):
            idx_begin = idx * batch_size
            idx_end = (idx + 1) * batch_size
            A = self.eval(x[idx_begin:idx_end, :], y)
            mat_vec_prod[idx_begin:idx_end] = A @ v
        return mat_vec_prod

    @abstractmethod
    def diagonal(self, X):
        pass

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def set_params(self, params):
        pass


# Abstract RBF
class L_RBF(Kernel):
    @abstractmethod
    def __init__(self):
        super(L_RBF, self).__init__()

    def eval(self, x, y):
        return self.rbf(self.ep, distance_matrix(np.atleast_2d(x), np.atleast_2d(y)))

    def diagonal(self, X):
        return np.ones(X.shape[0]) * self.rbf(self.ep, 0)

    def __str__(self):
        return self.name + ' [gamma = %2.2e]' % self.ep

    def set_params(self, par):
        self.ep = par

    def dd_diagonal(self, X):
        # Evaluation of (L^x - shift) (L^y - shift) k(x,y) for x=y: Likely some constant

        return np.ones(X.shape[0]) * self.dd_eval(np.zeros((1, 1)), np.zeros((1, 1))).item()

    def dd_eval(self, x, y):
        # Evaluation of (L^x - shift) (L^y - shift) k(x,y)

        array_dists = distance_matrix(np.atleast_2d(x), np.atleast_2d(y))

        return self.rbf_utility(self.ep, array_dists) \
               - 2 * self.shift * self.d2_eval(x, y) - self.shift ** 2 * self.rbf(self.ep, array_dists)

    def d2_eval(self, x, y):
        # Evaluation of (L^z - shift * Id) k(*,z)|_{x,y}
        # Required exactly like this for "compute the nth basis", see also 14.12.21\A1

        array_dists = distance_matrix(np.atleast_2d(x), np.atleast_2d(y))

        array_result = - ((self.dim - 1) * self.rbf1_divided_by_r(self.ep, array_dists)
                          + self.rbf2(self.ep, array_dists)) \
                       - self.shift * self.rbf(self.ep, array_dists)

        return array_result

    def d1_eval(self, x, y):
        # Evaluation of (L - shift * Id) k(*,y)|_{x}
        # Required exactly like this for "compute the nth basis", see also 14.12.21\A1

        # Since the LaPlace is radial, the formula is the same as d2_eval
        return self.d2_eval(x, y)

    def mixed_L_eval(self, x, ctrs, n_ctrs, list_ind_L, list_ind_k, list_ind_n):
        # L evaluations in x of Riesz representers which are located in ctrs.
        # list_ind_L, list_ind_k and list_ind_n indicate to which set of functionals the Riesz
        # representer belong to:
        # y1 := y[list_ind_L, :] corresponds to the interior, y2 := y[list_ind_k, :] to Dirichlet
        # boundary centers and y3 := y[list_ind_n, :] to Neumann boundary centers.
        # Different eval-methods need to be used with respect to y1, y2, y3.

        x = np.atleast_2d(x)
        ctrs = np.atleast_2d(ctrs)

        array_result = np.zeros((x.shape[0], ctrs.shape[0]))

        array_result[:, list_ind_L] = self.dd_eval(x, ctrs[list_ind_L])
        array_result[:, list_ind_k] = self.d1_eval(x, ctrs[list_ind_k])
        array_result[:, list_ind_n] = self.mixed_L_n(x, ctrs[list_ind_n], n_ctrs[list_ind_n])

        return array_result

    def mixed_k_eval(self, x, ctrs, n_ctrs, list_ind_L, list_ind_k, list_ind_n):
        # Point evaluations in x of Riesz representers which are located in ctrs.
        # list_ind_L, list_ind_k and list_ind_n indicate to which set of functionals the Riesz
        # representer belong to:
        # y1 := y[list_ind_L, :] corresponds to the interior, y2 := y[list_ind_k, :] to Dirichlet
        # boundary centers and y3 := y[list_ind_n, :] to Neumann boundary centers.
        # Different eval-methods need to be used with respect to y1, y2, y3.

        x = np.atleast_2d(x)
        ctrs = np.atleast_2d(ctrs)

        array_result = np.zeros((x.shape[0], ctrs.shape[0]))

        array_result[:, list_ind_L] = self.d2_eval(x, ctrs[list_ind_L])
        array_result[:, list_ind_k] = self.eval(x, ctrs[list_ind_k])
        array_result[:, list_ind_n] = self.mixed_k_n(x, ctrs[list_ind_n], n_ctrs[list_ind_n])

        return array_result

    def mixed_n_eval(self, x, n_x, ctrs, n_ctrs, list_ind_L, list_ind_k, list_ind_n):
        # Neumann derivative evaluations in x with normal vector n_x of Riesz
        # representers which are located in ctrs. n_ctrs contains normal vectors
        # to those centers listed in list_ind_n - if this list is empty, then the
        # array n_ctrs should be a zero array.
        # list_ind_L, list_ind_k and list_ind_n indicate to which set of functionals the Riesz
        # representer belong to:
        # y1 := y[list_ind_L, :] corresponds to the interior, y2 := y[list_ind_k, :] to Dirichlet
        # boundary centers and y3 := y[list_ind_n, :] to Neumann boundary centers.
        # Different eval-methods need to be used with respect to y1, y2, y3.

        x = np.atleast_2d(x)
        ctrs = np.atleast_2d(ctrs)

        array_result = np.zeros((x.shape[0], ctrs.shape[0]))

        array_result[:, list_ind_L] = self.mixed_L_n(ctrs[list_ind_L], x, n_x).T
        array_result[:, list_ind_k] = self.mixed_k_n(ctrs[list_ind_k], x, n_x).T
        array_result[:, list_ind_n] = self.mixed_n_n(
            x, ctrs[list_ind_n], n_x, n_ctrs[list_ind_n])

        return array_result

    # Now implement some functions related to Neumann boundary conditions
    def nn_diagonal(self, X):
        # Evaluation of < d/dn k(*, x), d/dn k(*, y) > for x = y,
        # i.e. dot product between two Neumann boundary functionals.
        # Value is independent of direction on n, therefore it is not provided/required.

        X = np.atleast_2d(X)

        return - np.ones(X.shape[0]) * self.rbf1_divided_by_r(self.ep, 0)

    def mixed_L_n(self, x, y, n_y):
        # Evaluation of < L^{2}k(*, x), d/dn k(*, y) >,  i.e. dot product of L evaluation and Neumann boundary
        # See 14.04.22\C, especially \C4.
        # n_y is the normal vectors for y.

        x = np.atleast_2d(x)
        y = np.atleast_2d(y)
        n_y = np.atleast_2d(n_y)

        array_dists = distance_matrix(x, y)

        array_result1 = x @ n_y.T - np.sum(y * n_y, axis=1, keepdims=True).T
        array_result2 = (self.dim - 1) * self.rbf_util_neuman(self.ep, array_dists) \
                        + self.rbf3_divided_by_r(self.ep, array_dists)

        return array_result1 * array_result2

    def mixed_k_n(self, x, y, n_y):
        # Evaluation of < k(*, x), d/dn k(*, y) >,  i.e. dot product of k evaluation and Neumann boundary
        # See 14.04.22\C, especially \C2.
        # n_y is the normal vectors for y.

        x = np.atleast_2d(x)
        y = np.atleast_2d(y)
        n_y = np.atleast_2d(n_y)

        array_dists = distance_matrix(x, y)

        array_result1 = np.sum(y * n_y, axis=1, keepdims=True).T - x @ n_y.T
        array_result2 = self.rbf1_divided_by_r(self.ep, array_dists)

        return array_result1 * array_result2

    def mixed_n_n(self, x, y, n_x, n_y):
        # Evaluation of < d/dn k(*, x), d/dn k(*, y) >,  i.e. dot product between two Neumann boundary functionals
        # See 14.04.22\C, especially \C2.
        # n_y is the normal vectors for y, same for n_x.

        x = np.atleast_2d(x)
        y = np.atleast_2d(y)
        n_x = np.atleast_2d(n_x)
        n_y = np.atleast_2d(n_y)

        array_dists = distance_matrix(x, y)

        array_1 = - self.rbf1_divided_by_r(self.ep, array_dists) * (n_x @ n_y.T)

        array_2 = - self.rbf_util_neuman(self.ep, array_dists) * \
                  (x @ n_y.T - np.sum(y * n_y, axis=1, keepdims=True).T) \
                  * (np.sum(x * n_x, axis=1, keepdims=True) - n_x @ y.T)

        return array_1 + array_2


# Implementation of concrete RBFs
class Gaussian_laplace(L_RBF):
    # # Gaussian kernel phi(r) = exp(-r^2),
    # used in conjunction with L = -LaPlace - shift * Id, whereby shift is some constant
    # (e.g. an eigenvalue of -LaPlace).

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1     # ToDo: Generalize this one!!!
        self.name = 'gauss'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.exp(-r**2)

        self.rbf1 = lambda ep, r: -2 * r * np.exp(-r**2)
        self.rbf1_divided_by_r = lambda ep, r: -2 * np.exp(-r**2)

        self.rbf2 = lambda ep, r: (4 * r ** 2 - 2) * np.exp(-r**2)

        self.rbf3_divided_by_r = lambda ep, r: -4 * (2 * r**2 - 3) * np.exp(-r**2)

        # Utility RBF for dd_eval: Double LaPlacian
        self.rbf_utility = lambda ep, r: \
                np.exp(-r ** 2) * (16 * r ** 4 - (16 * self.dim + 32) * r ** 2 + (4 * self.dim ** 2 + 8 * self.dim))

        # Utility RBF for neumann: See 14.04.22\C4, this is Phi'' / r^2 - Phi' / r^3
        self.rbf_util_neuman = lambda ep, r: 4 * np.exp(-r ** 2)


class wendland32_laplace(L_RBF):
    # Wendland kernel 3, 2: phi(r) = (1-r)_+^6 * (35r**2 + 18r + 3)
    # used in conjunction with L = -LaPlace - shift * Id, whereby shift is some constant
    # (e.g. an eigenvalue of -LaPlace).

    # Only for dim <= 3!

    def  __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1 # ep ToDo: Implement general case!
        self.name = 'Wendland 3, 2'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: 1/3 * np.maximum(1-r, 0)**6 * (35 * r**2 + 18 * r + 3)

        self.rbf1 = lambda ep, r: -1/3 * np.maximum(1 - r, 0) ** 5 * 56 * r * (5 * r + 1)
        self.rbf1_divided_by_r = lambda ep, r: -1/3 * np.maximum(1 - r, 0) ** 5 * 56 * (5 * r + 1)

        self.rbf2 = lambda ep, r: 1/3 * np.maximum(1 - r, 0) ** 4 * 56 * (35 * r ** 2 - 4 * r - 1)

        # Utility RBF for dd_eval: Double LaPlacian
        self.rbf_utility = lambda ep, r: \
            1/3 * (1680*self.dim**2 + 3360*self.dim
                   + r**4*(1680*self.dim**2 + 16800*self.dim + 40320)
                   + r**3*(-6720*self.dim**2 - 53760*self.dim - 100800)
                   + r**2*(10080*self.dim**2 + 60480*self.dim + 80640)
                   + r*(-6720*self.dim**2 - 26880*self.dim - 20160)) * (r < 1)

        # Utility RBF for neumann: See 14.04.22\C4, this is Phi'' / r^2 - Phi' / r^3
        self.rbf3_divided_by_r = lambda ep, r: np.zeros_like(r)     # ToDo: Not (yet) implemented
        self.rbf_util_neuman = lambda ep, r: np.zeros_like(r)       # ToDo: Not (yet) implemented


class wendland33_laplace(L_RBF):
    # Wendland kernel 3, 3: phi(r) = (1-r)_+^8 * (32r**3+ 25r^2 + 8r + 1)
    # used in conjunction with L = -LaPlace - shift * Id, whereby shift is some constant
    # (e.g. an eigenvalue of -LaPlace).

    # Only for dim <= 3!

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1  # ep ToDo: Implement general case!
        self.name = 'wendland 3, 3'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.maximum(1 - r, 0) ** 8 * (32 * r ** 3 + 25 * r ** 2 + 8 * r + 1)

        self.rbf1 = lambda ep, r: - np.maximum(1 - r, 0) ** 7 * 22 * r * (16 * r ** 2 + 7 * r + 1)
        self.rbf1_divided_by_r = lambda ep, r: - np.maximum(1 - r, 0) ** 7 * 22 * (16 * r ** 2 + 7 * r + 1)

        self.rbf2 = lambda ep, r: np.maximum(1 - r, 0) ** 6 * 22 * (160 * r ** 3 + 15 * r ** 2 - 6 * r - 1)

        self.rbf3_divided_by_r = lambda ep, r: -np.maximum(1-r, 0) ** 5 * 1584 * (20 * r ** 2 - 5 * r - 1)

        # Utility RBF for dd_eval: Double LaPlacian
        self.rbf_utility = lambda ep, r: \
            (528 * self.dim ** 2 + 1056 * self.dim + r ** 7 * (3168 * self.dim ** 2 + 50688 * self.dim + 199584)
             + r ** 6 * (-18480 * self.dim ** 2 - 258720 * self.dim - 887040)
             + r ** 5 * (44352 * self.dim ** 2 + 532224 * self.dim + 1552320)
             + r ** 4 * (-55440 * self.dim ** 2 - 554400 * self.dim - 1330560)
             + r ** 3 * (36960 * self.dim ** 2 + 295680 * self.dim + 554400)
             + r ** 2 * (-11088 * self.dim ** 2 - 66528 * self.dim - 88704)) * (r < 1)

        self.rbf_util_neuman = lambda ep, r: 528 * np.maximum(1 - r, 0) ** 6 * (6*r + 1)




class quadrMatern_laplace(L_RBF):
    # Quadratic matern kernel phi(r) = (3 + 3r + r^2)*exp(-r),
    # used in conjunction with L = -LaPlace

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1  # ep ToDo: Implement general case!
        self.name = 'quadratic Matern'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.exp(-r) * (3 + 3 *r + r ** 2)

        self.rbf1 = lambda ep, r: np.exp(-r) * (-r) * (r+1)
        self.rbf1_divided_by_r = lambda ep, r: np.exp(-r) * (-1) * (r+1)

        self.rbf2 = lambda ep, r: np.exp(-r) * (r ** 2 - r - 1)

        # Utility RBF for dd_eval: Double LaPlacian
        self.rbf_utility = lambda ep, r: \
            np.exp(-r) * (r**2 - (2 * self.dim + 3) * r + self.dim ** 2 + 2 * self.dim)

        # Utility RBF for neumann: See 14.04.22\C4, this is Phi'' / r^2 - Phi' / r^3
        self.rbf3_divided_by_r = lambda ep, r: np.zeros_like(r)  # ToDo: Not (yet) implemented
        self.rbf_util_neuman = lambda ep, r: np.zeros_like(r)       # ToDo: Not (yet) implemented


class cubicMatern_laplace(L_RBF):
    # Quadratic matern kernel phi(r) = (3 + 3r + r^2)*exp(-r),
    # used in conjunction with L = -LaPlace

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1  # ep ToDo: Implement general case!
        self.name = 'cubic Matern'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.exp(-r) * (15 + 15 * r + 6 * r ** 2 + r ** 3)

        self.rbf1 = lambda ep, r: np.exp(-r) * (-r) * (r ** 2 + 3 * r + 3)
        self.rbf1_divided_by_r = lambda ep, r: np.exp(-r) * (-1) * (r ** 2 + 3 * r + 3)

        self.rbf2 = lambda ep, r: np.exp(-r) * (r ** 3 - 3 * r - 3)

        # Utility RBF for dd_eval: Double LaPlacian
        self.rbf_utility = lambda ep, r: \
            np.exp(-r) * ((self.dim ** 2 + 2 * self.dim)
                          + (self.dim ** 2 + 2 * self.dim) * r
                          - (4 + 2 * self.dim) * r ** 2 + r ** 3)


        # Utility RBF for neumann: See 14.04.22\C4, this is Phi'' / r^2 - Phi' / r^3
        self.rbf3_divided_by_r = lambda ep, r: np.zeros_like(r)  # ToDo: Not (yet) implemented
        self.rbf_util_neuman = lambda ep, r: np.zeros_like(r)       # ToDo: Not (yet) implemented




class L_RBF_derivative(Kernel):
    # L = d/dx, i.e. only first derivative

    @abstractmethod
    def __init__(self):
        super(L_RBF_derivative, self).__init__()

    def eval(self, x, y):
        return self.rbf(self.ep, distance_matrix(np.atleast_2d(x), np.atleast_2d(y)))

    def diagonal(self, X):
        return np.ones(X.shape[0]) * self.rbf(self.ep, 0)

    def __str__(self):
        return self.name + ' [gamma = %2.2e]' % self.ep

    def set_params(self, par):
        self.ep = par

    def dd_diagonal(self, X):
        # Evaluation of (L^x - shift) (L^y - shift) k(x,y) for x=y: Likely some constant

        return np.ones(X.shape[0]) * self.dd_eval(np.zeros((1, 1)), np.zeros((1, 1))).item()

    def dd_eval(self, x, y):
        # Evaluation of (L^x - shift) (L^y - shift) k(x,y)

        array_dists = distance_matrix(np.atleast_2d(x), np.atleast_2d(y))

        return - self.rbf2(self.ep, array_dists)

    def d2_eval(self, x, y):
        # Evaluation of (L^z - shift * Id) k(*,z)|_{x,y}
        # Required exactly like this for "compute the nth basis", see also 14.12.21\A1

        array_dists = distance_matrix(np.atleast_2d(x), np.atleast_2d(y))
        array_x_minus_y = np.atleast_2d(x) - np.atleast_2d(y).transpose()

        # Introduce minus since derivative wrt y instead of x
        array_result = - np.sign(array_x_minus_y) * self.rbf1(self.ep, array_dists)

        return array_result

    def d1_eval(self, x, y):
        # Evaluation of (L - shift * Id) k(*,y)|_{x}
        # Required exactly like this for "compute the nth basis", see also 14.12.21\A1

        # Need a minus here since first derivative depends (in sign) on whether x or y
        return - self.d2_eval(x, y)

    def mixed_L_eval(self, x, ctrs, n_ctrs, list_ind_L, list_ind_k, list_ind_n):
        # L evaluations in x of Riesz representers which are located in ctrs.
        # list_ind_L, list_ind_k and list_ind_n indicate to which set of functionals the Riesz
        # representer belong to:
        # y1 := y[list_ind_L, :] corresponds to the interior, y2 := y[list_ind_k, :] to Dirichlet
        # boundary centers and y3 := y[list_ind_n, :] to Neumann boundary centers.
        # Different eval-methods need to be used with respect to y1, y2, y3.

        x = np.atleast_2d(x)
        ctrs = np.atleast_2d(ctrs)

        array_result = np.zeros((x.shape[0], ctrs.shape[0]))

        array_result[:, list_ind_L] = self.dd_eval(x, ctrs[list_ind_L])
        array_result[:, list_ind_k] = self.d1_eval(x, ctrs[list_ind_k])
        array_result[:, list_ind_n] = self.mixed_L_n(x, ctrs[list_ind_n], n_ctrs[list_ind_n])

        return array_result

    def mixed_k_eval(self, x, ctrs, n_ctrs, list_ind_L, list_ind_k, list_ind_n):
        # Point evaluations in x of Riesz representers which are located in ctrs.
        # list_ind_L, list_ind_k and list_ind_n indicate to which set of functionals the Riesz
        # representer belong to:
        # y1 := y[list_ind_L, :] corresponds to the interior, y2 := y[list_ind_k, :] to Dirichlet
        # boundary centers and y3 := y[list_ind_n, :] to Neumann boundary centers.
        # Different eval-methods need to be used with respect to y1, y2, y3.

        x = np.atleast_2d(x)
        ctrs = np.atleast_2d(ctrs)

        array_result = np.zeros((x.shape[0], ctrs.shape[0]))

        array_result[:, list_ind_L] = self.d2_eval(x, ctrs[list_ind_L])
        array_result[:, list_ind_k] = self.eval(x, ctrs[list_ind_k])
        array_result[:, list_ind_n] = self.mixed_k_n(x, ctrs[list_ind_n], n_ctrs[list_ind_n])

        return array_result

    def mixed_n_eval(self, x, n_x, ctrs, n_ctrs, list_ind_L, list_ind_k, list_ind_n):
        # not required therefore just return zeros

        return np.zeros((x.shape[0], ctrs.shape[0]))

    def nn_diagonal(self, X):
        # not required therefore just return zeros

        return np.zeros(X.shape[0])

    def mixed_L_n(self, x, y, n_y):
        # not required therefore just return zeros

        return np.zeros((x.shape[0], y.shape[0]))

    def mixed_k_n(self, x, y, n_y):
        # not required, therefore just return zeros

        return np.zeros((x.shape[0], y.shape[0]))

    def mixed_n_n(self, x, y, n_x, n_y):
        # not required, therefore just return zeros

        return np.zeros((x.shape[0], y.shape[0]))


class Gaussian_derivative(L_RBF_derivative):
    # # Gaussian kernel phi(r) = exp(-r^2),
    # used in conjunction with L = d/dx in 1D

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1     # ToDo: Generalize this one!!!
        self.name = 'gauss'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.exp(-r**2)
        self.rbf1 = lambda ep, r: -2 * r * np.exp(-r**2)
        self.rbf2 = lambda ep, r: (4 * r ** 2 - 2) * np.exp(-r**2)


class LinearMatern_derivative(L_RBF_derivative):
    # used in conjunction with L = d/dx in 1D

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1     # ToDo: Generalize this one!!!
        self.name = 'quadratic Matern'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r:  np.exp(-r) * (1 + r)
        self.rbf1 = lambda ep, r: - np.exp(-r) * r
        self.rbf2 = lambda ep, r: np.exp(-r) * (r - 1)

class QuadraticMatern_derivative(L_RBF_derivative):
    # used in conjunction with L = d/dx in 1D

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1     # ToDo: Generalize this one!!!
        self.name = 'quadratic Matern'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.exp(-r) * (3 + 3 * r + r ** 2)
        self.rbf1 = lambda ep, r: np.exp(-r) * (-r) * (r+1)
        self.rbf2 = lambda ep, r: np.exp(-r) * (r ** 2 - r - 1)



class CubicMatern_derivative(L_RBF_derivative):
    # used in conjunction with L = d/dx in 1D

    def __init__(self, dim, shift=0, ep=1):
        super().__init__()
        self.dim = dim
        self.shift = shift
        self.ep = 1     # ToDo: Generalize this one!!!
        self.name = 'cubic Matern'

        # Implement not only the rbf, but also its derivatives
        self.rbf = lambda ep, r: np.exp(-r) * (15 + 15 * r + 6 * r ** 2 + r ** 3)
        self.rbf1 = lambda ep, r: np.exp(-r) * (-r) * (r ** 2 + 3 * r + 3)
        self.rbf2 = lambda ep, r: np.exp(-r) * (r ** 3 - 3 * r - 3)



