# First approach to develope code for greedy PDE.
# ToDo: Not sure whether vector valued output makes sense, but let's try to implement it ..
# ToDo: Remove reg_par !! Does not really make sense for PDE approximation, the RHS is exact!

from vkoga_pde.kernels_PDE import wendland32_laplace
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array


# VKOGA implementation
class VKOGA_PDE(BaseEstimator):

    def __init__(self, kernel=wendland32_laplace(dim=2), kernel_par=1,
                 verbose=True, n_report=10,
                 greedy_type='p_greedy', reg_par=0, restr_par=0,
                 tol_f=1e-10, tol_p=1e-10, beta=None, weight_dirichlet=1):

        self.weight_dirichlet = weight_dirichlet

        # Set the verbosity on/off
        self.verbose = verbose

        # Set the frequency of report
        self.n_report = n_report

        # Set the params defining the method
        self.kernel = kernel
        self.kernel_par = kernel_par
        self.reg_par = reg_par
        self.restr_par = restr_par

        # Define the greedy method properly
        if beta is None:
            self.greedy_type = greedy_type
            if self.greedy_type == 'f_greedy':
                self.beta = 1.0
            elif self.greedy_type == 'p_greedy':
                self.beta = 0.0
            elif self.greedy_type == 'random':
                self.beta = 1.0         # use any value, should not matter!
        else:
            self.beta = float(beta)
            if self.beta == 1.0:
                self.greedy_type = 'f_greedy'
            elif self.beta == 0.0:
                self.greedy_type = 'p_greedy'
            else:
                self.greedy_type = '{}'.format(beta)

        # ToDo: Did never think about the regularization parameter ..
        assert reg_par == 0, 'reg_par > 0 might not be implemented correctly!'

        self.flag_val = None

        # Set the stopping values
        self.tol_f = tol_f
        self.tol_p = tol_p

        # Some further settings
        self.ctrs_ = None  # centers for the interior
        self.ind_ctrs1 = []  # self.ctrs1_ = self.ctrs_[self.indI_ctrs1], i.e. interior
        self.ind_ctrs2 = []  # self.ctrs2_ = self.ctrs_[self.indI_ctrs2], i.e. boundary Dirichlet
        self.ind_ctrs3 = []  # self.ctrs3_ = self.ctrs_[self.indI_ctrs3], i.e. boundary Neuman

        self.Cut_ = None
        self.c = None

        # Initialize the convergence history
        self.train_hist = {}
        self.train_hist['n'] = []
        self.train_hist['f'] = []           # max residual (squared)
        self.train_hist['f sol'] = []       # max residual (non-squared)
        self.train_hist['f1'] = []          # max residual inside with L
        self.train_hist['f2'] = []          # max residual boundary Dirichlet
        self.train_hist['f3'] = []          # max residual boundary Neumann
        self.train_hist['p'] = []
        self.train_hist['p1'] = []          # max Power function inside with L
        self.train_hist['p2'] = []          # max Power function boundary Dirichlet
        self.train_hist['p3'] = []          # max Power function boundary Neumann
        self.train_hist['p2_interior'] = []
        self.train_hist['p selected'] = []  # list of selected power vals
        self.train_hist['f val'] = []
        self.train_hist['p val'] = []

        # Initialize variables for the maximal kernel diagonal value in the interior and the boundary
        self.max_kernel_diag1 = None
        self.max_kernel_diag2 = None
        self.max_kernel_diag3 = None

        self.flag_neumann = None

    def selection_rule(self, f1, p1, f2, p2, f3, p3, flag_select_dirichlet=False):
        # Generalization of selection rule: We have residual values and power function values
        # for both the interior (f1, p1) as well as for the boundary (Dirichlet + Neumann) (f2, f3, p2, p3).

        # if self.restr_par > 0:
        #     # ToDo: Not sure whether this will really work, restr_idx might be empty
        #
        #     p_ = max([np.max(p1), np.max(p2), np.max(p3)])
        #     restr_idx1 = np.nonzero(p1 >= self.restr_par * p_)[0]
        #     restr_idx2 = np.nonzero(p2 >= self.restr_par * p_)[0]
        #     restr_idx3 = np.nonzero(p3 >= self.restr_par * p_)[0]
        # else:
        restr_idx1 = np.arange(len(p1))
        restr_idx2 = np.arange(len(p2))
        restr_idx3 = np.arange(len(p3))

        # Ensure that power function does not have negative values
        p1 = np.maximum(p1, 0, p1)
        p2 = np.maximum(p2, 0, p2)
        p3 = np.maximum(p3, 0, p3)

        f1 = np.sum(f1 ** 2, axis=1)
        f2 = np.sum(f2 ** 2, axis=1)
        f3 = np.sum(f3 ** 2, axis=1)

        if flag_select_dirichlet:
            f1 = 0 * f1
            f3 = 0 * f3

        # Case distinction required for fp_greedy
        if self.greedy_type == 'fp_greedy':
            # Extra implementation for f/P greedy because it's a limit case

            # Points in the interior
            if len(f1[restr_idx1]) != 0:    
                array_weighted1 = f1[restr_idx1] / p1[restr_idx1]
                idx1 = np.argmax(array_weighted1)
                array_weighted1_max = array_weighted1[idx1]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted1_max = 0.0

            # Dirichlet boundary
            if len(f2[restr_idx2]) != 0:  # required especially in 1D
                array_weighted2 = f2[restr_idx2] / p2[restr_idx2]
                idx2 = np.argmax(array_weighted2)
                array_weighted2_max = array_weighted2[idx2]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted2_max = 0.0

            # Neumann boundary
            if len(f3[restr_idx3]) != 0:  # required especially in 1D
                array_weighted3 = f3[restr_idx3] / p3[restr_idx3]
                idx3 = np.argmax(array_weighted3)
                array_weighted3_max = array_weighted3[idx3]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted3_max = 0.0

        elif self.greedy_type == 'p_greedy':
            # Extra implementation for P-greedy to address weighting of powerfunc (interior, boundary)

            # Points in the interior
            if len(f1[restr_idx1]) != 0:    
                array_weighted1 = p1[restr_idx1] / self.max_kernel_diag1
                idx1 = np.argmax(array_weighted1)
                array_weighted1_max = array_weighted1[idx1]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted1_max = 0.0

            # Dirichlet boundary
            if len(f2[restr_idx2]) != 0:  # required especially in 1D
                array_weighted2 = p2[restr_idx2] / self.max_kernel_diag2
                idx2 = np.argmax(array_weighted2)
                array_weighted2_max = self.weight_dirichlet * array_weighted2[idx2]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted2_max = 0.0

            # Neumann boundary
            if len(f3[restr_idx3]) != 0:  # required especially in 1D
                array_weighted3 = p3[restr_idx3] / self.max_kernel_diag3
                idx3 = np.argmax(array_weighted3)
                array_weighted3_max = array_weighted3[idx3]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted3_max = 0.0

        elif self.greedy_type == 'random':
            # Copied from below (beta-greedy) and argmax replaced by randint

            if len(f1[restr_idx1]) != 0:   
                array_weighted1 = f1[restr_idx1] ** self.beta * p1[restr_idx1] ** (1 - self.beta)
                # idx1 = np.argmax(array_weighted1)
                idx1 = np.random.randint(len(array_weighted1))
                array_weighted1_max = array_weighted1[idx1]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted1_max = 0.0

            # Dirichlet boundary
            if len(f2[restr_idx2]) != 0:  # required especially in 1D
                array_weighted2 = f2[restr_idx2] ** self.beta * p2[restr_idx2] ** (1 - self.beta)
                # idx2 = np.argmax(array_weighted2)
                idx2 = np.random.randint(len(array_weighted2))
                array_weighted2_max = self.weight_dirichlet * array_weighted2[idx2]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted2_max = 0.0

            # Neumann boundary
            if len(f3[restr_idx3]) != 0:  # required especially in 1D
                array_weighted3 = f3[restr_idx3] ** self.beta * p3[restr_idx3] ** (1 - self.beta)
                # idx3 = np.argmax(array_weighted3)
                idx3 = np.random.randint(len(array_weighted3))
                array_weighted3_max = array_weighted3[idx3]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted3_max = 0.0

        else:
            # Points in the interior
            if len(f1[restr_idx1]) != 0:    
                array_weighted1 = f1[restr_idx1] ** self.beta * p1[restr_idx1] ** (1 - self.beta)
                idx1 = np.argmax(array_weighted1)
                array_weighted1_max = array_weighted1[idx1]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted1_max = 0.0

            # Dirichlet boundary
            if len(f2[restr_idx2]) != 0:  # required especially in 1D
                array_weighted2 = f2[restr_idx2] ** self.beta * p2[restr_idx2] ** (1 - self.beta)
                idx2 = np.argmax(array_weighted2)
                array_weighted2_max = self.weight_dirichlet * array_weighted2[idx2]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted2_max = 0.0

            # Neumann boundary
            if len(f3[restr_idx3]) != 0:  # required especially in 1D
                array_weighted3 = f3[restr_idx3] ** self.beta * p3[restr_idx3] ** (1 - self.beta)
                idx3 = np.argmax(array_weighted3)
                array_weighted3_max = array_weighted3[idx3]
            else:
                # Set value to zero, such that it is the smallest possible
                array_weighted3_max = 0.0

        # Check where maximum is attained and set correspondingly
        list_max = [1 * array_weighted1_max, 1 * array_weighted2_max, 1 * array_weighted3_max]
        one_two_or_three = np.argmax(np.array(list_max)) + 1      # 1 = interior, 2 = Dirichlet, 3 = Neumann

        if one_two_or_three == 1:
            idx = restr_idx1[idx1]
        elif one_two_or_three == 2:
            idx = restr_idx2[idx2]
        elif one_two_or_three == 3:
            idx = restr_idx3[idx3]

        # The following implementation is quite laborious
        if len(f1[restr_idx1]) != 0:
            if self.greedy_type == 'p_greedy':

                f_max1, p_max1 = np.max(f1), np.max(p1) / self.max_kernel_diag1
            else:
                f_max1, p_max1 = np.max(f1), np.max(p1)
        else:
            f_max1, p_max1 = 0.0, 0.0

        # Dirichlet boundary
        if len(f2[restr_idx2]) != 0:  # required especially in 1D
            if self.greedy_type == 'p_greedy':
                f_max2, p_max2 = np.max(f2), np.max(p2) / self.max_kernel_diag2
            else:
                f_max2, p_max2 = np.max(f2), np.max(p2)
        else:
            f_max2, p_max2 = 0.0, 0.0

        # Neumann boundary
        if len(f3[restr_idx3]) != 0:  # required especially in 1D
            if self.greedy_type == 'p_greedy':
                f_max3, p_max3 = np.max(f3), np.max(p3) / self.max_kernel_diag3
            else:
                f_max3, p_max3 = np.max(f3), np.max(p3)
        else:
            f_max3, p_max3 = 0., 0.0

        return (one_two_or_three, idx), f_max1, f_max2, f_max3, p_max1, p_max2, p_max3

    def fit(self, X1, y1, X2, y2, X3=None, y3=None, n_X3=None,
            X1_val=None, y1_val=None, X2_val=None, y2_val=None, X3_val=None, y3_val=None, n_X3_val=None,
            maxIter=100, y1_sol=None):
        # ToDo: Tell the user what to do in case there are either no Dirichlet or no Neumann
        # ToDo: boundary points.

        # Points in X1 are considered as interior points, points in X2 as Dirichlet boundary points
        # and points in X3 are Neumann boundary points (whereby n3 are the normalized normal vectors).
        # They can coincide (partly), but for X1 we use the L-evaluation, for X2 we use the point
        # evaluation and for X3 we use evaluation of the normal derivative.
        # y1_sol is optional and contains the solution values at the points X1 - this is usefull to
        # track the convergence to the true solution in case it is known.

        list_ctrs = []  # not that beautiful, but required to store the final centers

        # Check the datasets
        X1, y1 = check_X_y(X1, y1, multi_output=True)
        X2, y2 = check_X_y(X2, y2, multi_output=True)
        # ToDo: Not sure whether this work properly if either X2 or X3 is empty

        # Check the validation datasets (if provided, otherwise set to None)
        if X1_val is None or y1_val is None \
                or X2_val is None or y2_val is None\
                or X3_val is None or y3_val is None:
            X1_val = None
            y1_val = None
            X2_val = None
            y2_val = None
            X3_val = None
            y3_val = None
            self.flag_val = False
        else:
            self.flag_val = True
            X1_val, y1_val = check_X_y(X1_val, y2_val, multi_output=True)
            X2_val, y2_val = check_X_y(X2_val, y2_val, multi_output=True)
            X3_val, y3_val = check_X_y(X3_val, y3_val, multi_output=True)
            # We will assume in the following that X_val and X are disjoint

            if len(y1_val.shape) == 1:
                y1_val = y1_val[:, None]
            if len(y2_val.shape) == 1:
                y2_val = y2_val[:, None]
            if len(y3_val.shape) == 1:
                y3_val = y3_val[:, None]

        if X3 is None or y3 is None or n_X3 is None:
            X3 = np.zeros((0, X1.shape[1]))
            y3 = np.zeros((0, X1.shape[1]))
            n_X3 = np.zeros((0, X1.shape[1]))
            self.flag_neumann = False
        else:
            X3, y3 = check_X_y(X3, y3, multi_output=True)
            self.flag_neumann = True

        if y1_sol is not None:
            y1_sol = check_array(y1_sol)

        if len(y1.shape) == 1:
            y1 = y1[:, None]
        if len(y2.shape) == 1:
            y2 = y2[:, None]
        if len(y3.shape) == 1:
            y3 = y3[:, None]

        # Compute the maximal kernel diagonal value in the interior and the boundary
        self.max_kernel_diag1 = np.max(self.kernel.dd_diagonal(X1))
        self.max_kernel_diag2 = np.max(self.kernel.diagonal(X2))
        if self.flag_neumann:
            self.max_kernel_diag3 = np.max(self.kernel.nn_diagonal(X3))

        # Check whether already fitted
        if self.ctrs_ is None:
            self.ctrs_ = np.zeros((0, X1.shape[1]))
            self.Cut_ = np.zeros((0, 0))
            self.c = np.zeros((0, y1.shape[1]))
            self.n_ctrs_ = np.zeros((maxIter, X1.shape[1]))       # stores normal vectors (only for Neumann ctrs)
            # ToDo: If self.ctrs_ is not None, does this work?

        # Check whether "new X1" and previously chosen centers overlap
        list_truly_new_X1 = []
        if self.ctrs_.shape[0] > 0:
            for idx_x in range(X1.shape[0]):
                if min(np.linalg.norm(self.ctrs_ - X1[idx_x, :], axis=1)) < 1e-10:
                    continue
                else:
                    list_truly_new_X1.append(idx_x)
        else:
            list_truly_new_X1 = list(range(X1.shape[0]))
        X1 = X1[list_truly_new_X1, :]
        y1 = y1[list_truly_new_X1, :]

        # Check whether "new X2" and previously chosen centers overlap
        list_truly_new_X2 = []
        if self.ctrs_.shape[0] > 0:
            for idx_x in range(X2.shape[0]):
                if min(np.linalg.norm(self.ctrs_ - X2[idx_x, :], axis=1)) < 1e-10:
                    continue
                else:
                    list_truly_new_X2.append(idx_x)
        else:
            list_truly_new_X2 = list(range(X2.shape[0]))
        X2 = X2[list_truly_new_X2, :]
        y2 = y2[list_truly_new_X2, :]

        # Check whether "new X3" and previously chosen centers overlap
        if self.flag_neumann:
            list_truly_new_X3 = []
            if self.ctrs_.shape[0] > 0:
                for idx_x in range(X3.shape[0]):
                    if min(np.linalg.norm(self.ctrs_ - X3[idx_x, :], axis=1)) < 1e-10:
                        continue
                    else:
                        list_truly_new_X3.append(idx_x)
            else:
                list_truly_new_X3 = list(range(X3.shape[0]))
            X3 = X3[list_truly_new_X3, :]
            y3 = y3[list_truly_new_X3, :]

        # Initialize the residual and update the given y values by substracting the current model
        y1 = y1 - self.predict_Ls(X1)
        y2 = y2 - self.predict_s(X2)
        if self.flag_neumann:
            y3 = y3 - self.predict_ns(X3, n_X3)

        # if self.flag_val:
        #     y1_val = y1_val - self.predict_Ls(X1_val)
        #     y2_val = y2_val - self.predict_s(X2_val)
        #     y3_val = y3_val - self.predict_ns(X3_val, n_X3_val)

        # Get the data dimension
        N, q = y1.shape[0] + y2.shape[0] + y3.shape[0], y1.shape[1]
        N1, N2, N3 = y1.shape[0], y2.shape[0], y3.shape[0]
        # if self.flag_val:
        #     N_val = y1_val.shape[0] + y2_val.shape[0] + y3_val.shape[0]
        #     N1_val, N2_val, N3_val = y1_val.shape[0], y2_val.shape[0], y3_val.shape[0]

        # Set maxIter_continue
        if maxIter > N:
            self.maxIter = N
        else:
            self.maxIter = maxIter

        # Check compatibility of restriction
        if self.beta == 0.0:  # in case of P-greedy
            self.restr_par = 0
        if not self.reg_par == 0:
            self.restr_par = 0

        # Initialize list for the chosen and non-chosen indices - seperately for interior and boundary
        indI1_, indI2_, indI3_ = [], [], []
        notIndI1, notIndI2, notIndI3 = list(range(y1.shape[0])), list(range(y2.shape[0])), list(range(y3.shape[0]))
        c = np.zeros((self.maxIter, q))

        # Compute the Newton basis values (related to the old centers) on the new X
        Vx_new_X1_old_ctrs = self.kernel.mixed_k_eval(
            X1, self.ctrs_, self.n_ctrs_, self.ind_ctrs1, self.ind_ctrs2, self.ind_ctrs3) @ self.Cut_.transpose()
        Vx_new_X2_old_ctrs = self.kernel.mixed_k_eval(
            X2, self.ctrs_, self.n_ctrs_, self.ind_ctrs1, self.ind_ctrs2, self.ind_ctrs3) @ self.Cut_.transpose()
        # if self.flag_val:
        #     Vx_val_new_X1_old_ctrs = self.kernel.mixed_k_eval(
        #         X1_val, self.ctrs_, self.ind_ctrs1) @ self.Cut_.transpose()
        #     Vx_val_new_X2_old_ctrs = self.kernel.mixed_k_eval(
        #         X2_val, self.ctrs_, self.ind_ctrs1) @ self.Cut_.transpose()


        # Compute the application of L to the Newton basis values (related to the old centers) on the new X
        LVx_new_X1_old_ctrs = self.kernel.mixed_L_eval(
            X1, self.ctrs_, self.n_ctrs_, self.ind_ctrs1, self.ind_ctrs2, self.ind_ctrs3) @ self.Cut_.transpose()
        # if self.flag_val:
        #     LVx_val_new_X1_old_ctrs = self.kernel.mixed_L_eval(
        #         X1_val, self.ctrs_, self.ind_ctrs1) @ self.Cut_.transpose()
        #     LVx_val_new_X2_old_ctrs = self.kernel.mixed_L_eval(
        #         X2_val, self.ctrs_, self.ind_ctrs1) @ self.Cut_.transpose()

        # Compute the Neumann derivatives of the Newton basis values (related to the old centers) on the new X
        if self.flag_neumann:
            nVx_new_X3_old_ctrs = self.kernel.mixed_n_eval(
                X3, n_X3, self.ctrs_, self.n_ctrs_, self.ind_ctrs1, self.ind_ctrs2, self.ind_ctrs3) @ self.Cut_.transpose()
        else:
            nVx_new_X3_old_ctrs = np.zeros((0, self.ctrs_.shape[0]))


        # Initialize arrays for the (L-, n-evaluation of the) Newton basis values ..
        # .. (related to the new centers) on the new X1, X2
        Vx1 = np.zeros((N1, self.maxIter))
        Vx2 = np.zeros((N2, self.maxIter))

        LVx1 = np.zeros((N1, self.maxIter))
        nVx3 = np.zeros((N3, self.maxIter))

        # if self.flag_val:
        #     Vx1_val = np.zeros((N1_val, self.maxIter))
        #     Vx2_val = np.zeros((N2_val, self.maxIter))
        #     LVx1_val = np.zeros((N1_val, self.maxIter))
        #     LVx2_val = np.zeros((N2_val, self.maxIter))

        # Compute the powervals on X1, X2 and X1_val, X2_val
        # ToDo: Vielleicht quadrieren?
        p1 = self.kernel.dd_diagonal(X1) + self.reg_par
        p1 = p1 - np.sum(LVx_new_X1_old_ctrs ** 2, axis=1)

        p2 = self.kernel.diagonal(X2) + self.reg_par
        p2 = p2 - np.sum(Vx_new_X2_old_ctrs ** 2, axis=1)

        p3 = self.kernel.nn_diagonal(X3) + self.reg_par
        p3 = p3 - np.sum(nVx_new_X3_old_ctrs ** 2, axis=1)


        # The following is a newly implemented quantity which is interesting for 1D
        p2_interior = self.kernel.diagonal(X1) + self.reg_par
        p2_interior = p2_interior - np.sum(Vx_new_X1_old_ctrs ** 2, axis=1)

        # if self.flag_val:
        #     p1_val = self.kernel.dd_diagonal(X1_val) + self.reg_par
        #     p1_val = p1_val - np.sum(LVx_val_new_X1_old_ctrs ** 2, axis=1)
        #
        #     p2_val = self.kernel.diagonal(X2_val) + self.reg_par
        #     p2_val = p2_val - np.sum(Vx_val_new_X2_old_ctrs ** 2, axis=1)

        # Extend Cut_ matrix, i.e. continue to build on old self.Cut_ matrix
        N_ctrs_so_far, N_ctrs_so_far1, N_ctrs_so_far2, N_ctrs_so_far3 = \
            self.Cut_.shape[0], len(self.ind_ctrs1), len(self.ind_ctrs2), len(self.ind_ctrs3)

        Cut_ = np.zeros((N_ctrs_so_far + self.maxIter, N_ctrs_so_far + self.maxIter))
        Cut_[:N_ctrs_so_far, :N_ctrs_so_far] = self.Cut_

        # Iterative selection of new points
        self.print_message('begin')

        n1, n2, n3 = 0, 0, 0  # number of centers in interior or boundary (Dirichlet, Neumann)
        for n in range(self.maxIter):
            # prepare
            self.train_hist['n'].append(self.ctrs_.shape[0] + n + 1)

            self.train_hist['f sol'].append([])     # tracking of approx of solution (if solution values are provided)
            self.train_hist['f'].append([])         # max of f1, f2, f3 below
            self.train_hist['f1'].append([])
            self.train_hist['f2'].append([])
            self.train_hist['f3'].append([])

            self.train_hist['p'].append([])         # max of p1, p2, p3 below
            self.train_hist['p1'].append([])        # power function of L evaluation in interior
            self.train_hist['p2'].append([])        # power function of Dirichlet evaluation on boundary
            self.train_hist['p3'].append([])        # power function of Neumann evaluation on boundary
            self.train_hist['p2_interior'].append(np.max(p2_interior))      # novel tracking, interestind in 1D
            self.train_hist['p selected'].append([])        # selected power val from either p1, p2, p3
            # if self.flag_val:
            #     self.train_hist['p val'].append([])
            #     self.train_hist['f val'].append([])

            # select the current index
            flag_dirichlet = False      # quickly played around a little bit
            # if n < 40:
            #     flag_dirichlet = True
            # else:
            #     flag_dirichlet = False

            (one_two_or_three, idx), \
                self.train_hist['f1'][-1], self.train_hist['f2'][-1], self.train_hist['f3'][-1], \
                self.train_hist['p1'][-1], self.train_hist['p2'][-1], self.train_hist['p3'][-1] \
                = self.selection_rule(y1[notIndI1], p1[notIndI1], y2[notIndI2], p2[notIndI2], y3[notIndI3], p3[notIndI3],
                                      flag_select_dirichlet=flag_dirichlet)
            self.train_hist['f'][-1] = max(self.train_hist['f1'][-1], self.train_hist['f2'][-1], self.train_hist['f3'][-1])
            self.train_hist['p'][-1] = max(self.train_hist['p1'][-1], self.train_hist['p2'][-1], self.train_hist['p3'][-1])

            # Append residual value (compared to true solution which was provided optionally in y1_sol)
            if y1_sol is not None:
                if n == 0:          # zero prediction for empty model
                    y1_sol_pred = np.zeros_like(y1_sol)
                else:               # make use of Newton basis elements
                    y1_sol_pred = Vx1[:, :n] @ c[:n, :]

                self.train_hist['f sol'][-1] = np.max(np.abs(y1_sol - y1_sol_pred))


            # add the current index and store selected p value and store the selected center
            if one_two_or_three == 1:
                self.train_hist['p selected'][-1] = p1[notIndI1[idx]]
                indI1_.append(notIndI1[idx])

                self.ind_ctrs1.append(N_ctrs_so_far + n)
                list_ctrs.append(X1[[notIndI1[idx]], :])
            elif one_two_or_three == 2:
                self.train_hist['p selected'][-1] = p2[notIndI2[idx]]
                indI2_.append(notIndI2[idx])

                self.ind_ctrs2.append(N_ctrs_so_far + n)
                list_ctrs.append(X2[[notIndI2[idx]], :])
            elif one_two_or_three == 3:
                self.train_hist['p selected'][-1] = p3[notIndI3[idx]]
                indI3_.append(notIndI3[idx])

                self.ind_ctrs3.append(N_ctrs_so_far + n)
                list_ctrs.append(X3[[notIndI3[idx]], :])

                # Add normal vector
                # ToDo: This does not work correctly if there already exist chosen centers
                self.n_ctrs_[n, :] = n_X3[notIndI3[idx]]

            # if self.flag_val:
            #     self.train_hist['p val'][-1] = max([np.max(p1_val), np.max(p2_val)])
            #     self.train_hist['f val'][-1] = max(
            #         [np.max(np.sum(y1_val ** 2, axis=1)), np.max(np.sum(y2_val ** 2, axis=1))])

            # check if the tolerances are reacheded
            if self.train_hist['f'][n] <= self.tol_f or self.train_hist['p'][n] <= self.tol_p:
                n = n - 1
                if one_two_or_three == 1:
                    n1 = n1 - 1
                elif one_two_or_three == 2:
                    n2 = n2 - 1
                elif one_two_or_three == 3:
                    n3 = n3 - 1
                self.print_message('end')
                break

            # compute the nth basis (including normalization), see also 14.12.21\A1
            if one_two_or_three == 1:
                # Computations (this is simply expansion in Newton basis) ..
                Vx1[:, n] = self.kernel.d2_eval(X1[:, :], X1[indI1_[n1], :])[:, 0] \
                            - Vx_new_X1_old_ctrs[:, :] @ LVx_new_X1_old_ctrs[indI1_[n1], :].transpose() \
                            - Vx1[:, :] @ LVx1[indI1_[n1], :].transpose()
                Vx2[:, n] = self.kernel.d2_eval(X2[:, :], X1[indI1_[n1], :])[:, 0] \
                            - Vx_new_X2_old_ctrs[:, :] @ LVx_new_X1_old_ctrs[indI1_[n1], :].transpose() \
                            - Vx2[:, :] @ LVx1[indI1_[n1], :].transpose()

                LVx1[:, n] = self.kernel.dd_eval(X1[:, :], X1[indI1_[n1], :])[:, 0] \
                             - LVx_new_X1_old_ctrs[:, :] @ LVx_new_X1_old_ctrs[indI1_[n1], :].transpose() \
                             - LVx1[:, :] @ LVx1[indI1_[n1], :].transpose()

                nVx3[:, n] = self.kernel.mixed_L_n(X1[indI1_[n1], :], X3, n_X3)[0, :].transpose() \
                            - nVx_new_X3_old_ctrs[:, :] @ LVx_new_X1_old_ctrs[indI1_[n1], :].transpose() \
                            - nVx3[:, :] @ LVx1[indI1_[n1], :].transpose()

                # Normalizations ..
                Vx1[:, n] = Vx1[:, n] / np.sqrt(p1[indI1_[n1]])
                Vx2[:, n] = Vx2[:, n] / np.sqrt(p1[indI1_[n1]])
                LVx1[:, n] = LVx1[:, n] / np.sqrt(p1[indI1_[n1]])
                nVx3[:, n] = nVx3[:, n] / np.sqrt(p1[indI1_[n1]])

                # Calculations for validation sets

            elif one_two_or_three == 2:
                # Computations (this is simply expansion in Newton basis) ..
                Vx1[:, n] = self.kernel.eval(X1[:, :], X2[indI2_[n2], :])[:, 0] \
                            - Vx_new_X1_old_ctrs[:, :] @ Vx_new_X2_old_ctrs[indI2_[n2], :].transpose() \
                            - Vx1[:, :] @ Vx2[indI2_[n2], :].transpose()
                Vx2[:, n] = self.kernel.eval(X2[:, :], X2[indI2_[n2], :])[:, 0] \
                            - Vx_new_X2_old_ctrs[:, :] @ Vx_new_X2_old_ctrs[indI2_[n2], :].transpose() \
                            - Vx2[:, :] @ Vx2[indI2_[n2], :].transpose()

                LVx1[:, n] = self.kernel.d1_eval(X1[:, :], X2[indI2_[n2], :])[:, 0] \
                             - LVx_new_X1_old_ctrs[:, :] @ Vx_new_X2_old_ctrs[indI2_[n2], :].transpose() \
                             - LVx1[:, :] @ Vx2[indI2_[n2], :].transpose()

                nVx3[:, n] = self.kernel.mixed_k_n(X2[indI2_[n2], :], X3, n_X3)[0, :].transpose() \
                            - nVx_new_X3_old_ctrs[:, :] @ Vx_new_X2_old_ctrs[indI2_[n2], :].transpose() \
                            - nVx3[:, :] @ Vx2[indI2_[n2], :].transpose()

                # Normalizations ..
                if p2[indI2_[n2]] < 0:
                    asdf = 1+2
                    print('hallo')
                Vx1[:, n] = Vx1[:, n] / np.sqrt(p2[indI2_[n2]])
                Vx2[:, n] = Vx2[:, n] / np.sqrt(p2[indI2_[n2]])
                LVx1[:, n] = LVx1[:, n] / np.sqrt(p2[indI2_[n2]])
                nVx3[:, n] = nVx3[:, n] / np.sqrt(p2[indI2_[n2]])

                # Calculations for validation sets

            elif one_two_or_three == 3:
                # Computations (this is simply expansion in Newton basis) ..
                Vx1[:, n] = self.kernel.mixed_k_n(X1[:, :], X3[indI3_[n3], :], self.n_ctrs_[[n], :])[:, 0] \
                            - Vx1[:, :] @ nVx3[indI3_[n3], :].transpose()
                            # - Vx_new_X1_old_ctrs[:, :] @ nVx_new_X3_old_ctrs[indI3_[n3], :].transpose() \
                Vx2[:, n] = self.kernel.mixed_k_n(X2[:, :], X3[indI3_[n3], :], self.n_ctrs_[[n], :])[:, 0] \
                            - Vx2[:, :] @ nVx3[indI3_[n3], :].transpose()
                            # - Vx_new_X2_old_ctrs[:, :] @ nVx_new_X3_old_ctrs[indI3_[n3], :].transpose() \

                LVx1[:, n] = self.kernel.mixed_L_n(X1[:, :], X3[indI3_[n3], :], self.n_ctrs_[[n], :])[:, 0] \
                            - LVx1[:, :] @ nVx3[indI3_[n3], :].transpose()
                            # - LVx_new_X1_old_ctrs[:, :] @ nVx_new_X3_old_ctrs[indI3_[n3], :].transpose() \

                nVx3[:, n] = self.kernel.mixed_n_n(X3[[indI3_[n3]], :], X3, self.n_ctrs_[[n], :], n_X3)[0, :].transpose() \
                            - nVx3[:, :] @ nVx3[indI3_[n3], :].transpose()
                            # - nVx_new_X3_old_ctrs[:, :] @ nVx_new_X3_old_ctrs[indI3_[n3], :].transpose() \

                # Normalizations ..
                Vx1[:, n] = Vx1[:, n] / np.sqrt(p3[indI3_[n3]])
                Vx2[:, n] = Vx2[:, n] / np.sqrt(p3[indI3_[n3]])
                LVx1[:, n] = LVx1[:, n] / np.sqrt(p3[indI3_[n3]])
                nVx3[:, n] = nVx3[:, n] / np.sqrt(p3[indI3_[n3]])

                # Calculations for validation sets

            # update the change of basis (see 14.12.21\B)
            Cut_new_row = np.ones(N_ctrs_so_far + n + 1)  # required only for the final one

            if one_two_or_three == 1:
                Cut_new_row[:N_ctrs_so_far + n] = \
                    -np.concatenate((LVx_new_X1_old_ctrs[indI1_[n1], :], LVx1[indI1_[n1], :n])) \
                    @ Cut_[:N_ctrs_so_far + n, :N_ctrs_so_far + n]

                Cut_[N_ctrs_so_far + n, :N_ctrs_so_far + n + 1] = Cut_new_row / LVx1[indI1_[n1], n]
            elif one_two_or_three == 2:
                Cut_new_row[:N_ctrs_so_far + n] = \
                    -np.concatenate((Vx_new_X2_old_ctrs[indI2_[n2], :], Vx2[indI2_[n2], :n])) \
                    @ Cut_[:N_ctrs_so_far + n, :N_ctrs_so_far + n]

                Cut_[N_ctrs_so_far + n, :N_ctrs_so_far + n + 1] = Cut_new_row / Vx2[indI2_[n2], n]

            elif one_two_or_three == 3:
                Cut_new_row[:N_ctrs_so_far + n] = \
                    -np.concatenate((nVx_new_X3_old_ctrs[indI3_[n3], :], nVx3[indI3_[n3], :n])) \
                    @ Cut_[:N_ctrs_so_far + n, :N_ctrs_so_far + n]

                Cut_[N_ctrs_so_far + n, :N_ctrs_so_far + n + 1] = Cut_new_row / nVx3[indI3_[n3], n]

            # compute the nth coefficient
            if one_two_or_three == 1:
                c[n] = y1[indI1_[n1]] / np.sqrt(p1[indI1_[n1]])
            elif one_two_or_three == 2:
                c[n] = y2[indI2_[n2]] / np.sqrt(p2[indI2_[n2]])
            elif one_two_or_three == 3:
                c[n] = y3[indI3_[n3]] / np.sqrt(p3[indI3_[n3]])

            # update the power functions: ToDo: For Gaussian in 1D we obtain negative power vals :(
            p1[notIndI1] = p1[notIndI1] - LVx1[notIndI1, n] ** 2
            p2[notIndI2] = p2[notIndI2] - Vx2[notIndI2, n] ** 2

            if np.min(p2) < -0.0005:
                aasdfff = 1 + 2
                print('debug: min(p2), median(p2) = {:.3e}, {:.3f}'.format(np.min(p2), np.median(p2)))

            p3[notIndI3] = p3[notIndI3] - nVx3[notIndI3, n] ** 2
            p2_interior = p2_interior - Vx1[:, n] ** 2

            # Calculations for validation sets
            # if self.flag_val:
            #     p1_val = p1_val - LVx1_val[:, n] ** 2
            #     p2_val = p2_val - Vx2_val[:, n] ** 2

            # update the residual
            y1[notIndI1] = y1[notIndI1] - LVx1[notIndI1, n][:, None] * c[n]
            y2[notIndI2] = y2[notIndI2] - Vx2[notIndI2, n][:, None] * c[n]
            y3[notIndI3] = y3[notIndI3] - nVx3[notIndI3, n][:, None] * c[n]

            # Calculations for validation sets
            # if self.flag_val:
            #     y1_val = y1_val - LVx1_val[:, n][:, None] * c[n]
            #     y2_val = y2_val - Vx2_val[:, n][:, None] * c[n]

            # remove the nth index from the dictionary
            if one_two_or_three == 1:
                notIndI1.pop(idx)
                n1 += 1
            elif one_two_or_three == 2:
                notIndI2.pop(idx)
                n2 += 1
            elif one_two_or_three == 3:
                notIndI3.pop(idx)
                n3 += 1

            # Report some data every now and then
            if n % self.n_report == 0:
                self.print_message('track')

        else:
            self.print_message('end')

        # Define coefficients and centers
        # ToDo: Check whether everything is clipped here correctly!!
        self.c = np.concatenate((self.c, c[:n + 1]))
        self.Cut_ = Cut_[:N_ctrs_so_far + n + 1, :N_ctrs_so_far + n + 1]
        self.ind_ctrs1 = self.ind_ctrs1[:N_ctrs_so_far1 + n1 + 1]  # ToDo: Check whether this is correct!!!!
        self.ind_ctrs2 = self.ind_ctrs2[:N_ctrs_so_far2 + n2 + 1]
        self.ind_ctrs3 = self.ind_ctrs3[:N_ctrs_so_far3 + n3 + 1]
        self.coef_ = self.Cut_.transpose() @ self.c

        self.ctrs_ = np.concatenate((self.ctrs_,
                                     np.concatenate(list_ctrs[:n + 1], axis=0)),
                                    axis=0)

        self.Vx1 = Vx1
        self.Vx2 = Vx2
        self.LVx1 = LVx1
        self.nVx3 = nVx3

        return self

    def predict_s(self, X):
        # Prediction of s, which might/should approximate the true solution u

        X = check_array(X)

        # Compute prediction
        if self.ctrs_.shape[0] > 0:
            prediction = self.kernel.d2_eval(X, self.ctrs_[self.ind_ctrs1]) @ self.coef_[self.ind_ctrs1] \
                         + self.kernel.eval(X, self.ctrs_[self.ind_ctrs2]) @ self.coef_[self.ind_ctrs2] \
                         + self.kernel.mixed_k_n(X, self.ctrs_[self.ind_ctrs3], self.n_ctrs_[self.ind_ctrs3]) @ self.coef_[self.ind_ctrs3]
        else:
            prediction = np.zeros((X.shape[0], 1))

        return prediction

    def predict_Ls(self, X):
        # Prediction of Ls, i.e. application of L to the prediction

        # Validate the input
        X = check_array(X)

        # Compute predictionf
        if self.ctrs_.shape[0] > 0:
            prediction = self.kernel.dd_eval(X, self.ctrs_[self.ind_ctrs1]) @ self.coef_[self.ind_ctrs1] \
                         + self.kernel.d1_eval(X, self.ctrs_[self.ind_ctrs2]) @ self.coef_[self.ind_ctrs2] \
                         + self.kernel.mixed_L_n(X, self.ctrs_[self.ind_ctrs3], self.n_ctrs_[self.ind_ctrs3]) @ self.coef_[self.ind_ctrs3]
        else:
            prediction = np.zeros((X.shape[0], 1))

        return prediction

    def predict_ns(self, X, n_x):
        # Prediction of ns, i.e. evaluation of normal derivative on given points

        # Validate the input
        X = check_array(X)

        # Compute prediction
        if self.ctrs_.shape[0] > 0:
            prediction = self.kernel.mixed_L_n(self.ctrs_[self.ind_ctrs1], X, n_x) @ self.coef_[self.ind_ctrs1] \
                         + self.kernel.mixed_k_n(self.ctrs_[self.ind_ctrs2], X, n_x) @ self.coef_[self.ind_ctrs2] \
                         + self.kernel.mixed_n_n(X, n_x, self.ctrs_[self.ind_ctrs3], self.n_ctrs_[self.ind_ctrs3]) @ self.coef_[self.ind_ctrs3]
        else:
            prediction = np.zeros((X.shape[0], 1))

        return prediction

    def print_message(self, when):
        if self.verbose and when == 'begin':
            print('')
            print('*' * 30 + ' [PDE-VKOGA] ' + '*' * 30)
            print('Training model with')
            print('       |_ kernel              : %s' % self.kernel)
            print('       |_ regularization par. : %2.2e' % self.reg_par)
            print('       |_ restriction par.    : %2.2e' % self.restr_par)
            print('')

        if self.verbose and when == 'end':
            print('Training completed with')
            print('       |_ selected points     : %8d / %8d' % (
            self.train_hist['n'][-1], self.ctrs_.shape[0] + self.maxIter))
            if self.flag_val:
                print('       |_ train, val residual : %2.2e / %2.2e,    %2.2e' %
                      (self.train_hist['f'][-1], self.tol_f, self.train_hist['f val'][-1]))
                print('       |_ train, val power fun: %2.2e / %2.2e,    %2.2e' %
                      (self.train_hist['p'][-1], self.tol_p, self.train_hist['p val'][-1]))
            else:
                print('       |_ train residual      : %2.2e / %2.2e' % (self.train_hist['f'][-1], self.tol_f))
                print('       |_ train power fun     : %2.2e / %2.2e' % (self.train_hist['p'][-1], self.tol_p))

        if self.verbose and when == 'track':
            print('Training ongoing with')
            print('       |_ selected points     : %8d / %8d' % (
            self.train_hist['n'][-1], self.ctrs_.shape[0] + self.maxIter))
            if self.flag_val:
                print('       |_ train, val residual : %2.2e / %2.2e,    %2.2e' %
                      (self.train_hist['f'][-1], self.tol_f, self.train_hist['f val'][-1]))
                print('       |_ train, val power fun: %2.2e / %2.2e,    %2.2e' %
                      (self.train_hist['p'][-1], self.tol_p, self.train_hist['p val'][-1]))
            else:
                print('       |_ train residual      : %2.2e / %2.2e' % (self.train_hist['f'][-1], self.tol_f))
                print('       |_ train power fun     : %2.2e / %2.2e' % (self.train_hist['p'][-1], self.tol_p))


# %% Utilities to
import pickle


def save_object(obj, filename):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)


def load_object(filename):
    with open(filename, 'rb') as input:
        obj = pickle.load(input)
    return obj
